/*
%CC1 $1 -X -E5000
%CLINK $1 WILDEXP -S
%DELETE $1.CRL 
*/

/***************************************************************
				COPY.C
****************************************************************
	Written by Leor Zolman, 3/82
	Modified for v1.50 9/82, 3/83
	Modified by Gene Mallory 12/83
****************************************************************
	COPY copies files to and from different drives and user areas on
	a CP/M filesystem.

	Wildcard expansion is performed only on files present in the
	currently logged drive and user area (the "current directory").

	Any filename may contain an optional disk designator prefix of
	the form <d:> and/or a user area number prefix of the form <n/>
	preceding the actual name. If both must be given, the user area
	prefix must come first.

	When a single file is to be copied, simply give the old and new
	filenames.

	When multiple files are to be copied, the destination parameter
	must consist of either a disk designator and/or user area pathname,
	or the single character "." when the destination is the CURRENT
	drive and user area. Such a destination designation is also allowed
	when only a single file is being copied.

	For example, to copy *.COM from the current directory into user
	area 0 of drive C:

		copy  *.com  0/C:

	To copy FOO.C from user area 4 into the current user area:

		copy 4/foo.c .

*/

#INCLUDE <bdscio.h>

MAIN(argc,argv)
INT argc;
CHAR **argv;
  BEGIN
    INT i,j,k,c,loopcnt;
    INT fd1,fd2;
    INT bufsects;
    UNSIGNED bufsize;
    UNSIGNED corebuf;
    CHAR cur_disk;  	/* currently logged-in disk */
    CHAR destname[30];
    CHAR *lastarg;
    CHAR overwrite, verify, *vbuf;  	/* true when verifying writes */

    overwrite = verify = FALSE;

/*********************************************************************
***               ARGUMENT PROCESSING                              ***
*********************************************************************/

INT ii,jj,optionerr;
CHAR *ss;
	optionerr = 0;
	FOR (ii=argc-1;ii>0;ii--)
	    IF (argv[ii][0] == '-')
		BEGIN
		    FOR (ss = &argv[ii][1]; *ss != '\0';)
			BEGIN
			    SWITCH (toupper(*ss++))
				BEGIN
				    CASE 'V':
					verify = TRUE;
					BREAK;
				    CASE 'O':
					overwrite = TRUE;
					BREAK;
				    CASE 'H':
					optionerr = TRUE;
					BREAK;
				    DEFAULT:
					printf
                                          (
                                          "COPY: Illegal option %c.\n"
                                          ,*--ss
                                          );
					ss++;
					optionerr = TRUE;
					BREAK;
				END
			    WHILE (isdigit(*ss)) ss++;
			END
		    FOR (jj=ii;jj<(argc-1);jj++) argv[jj] = argv[jj+1];
		    argc--;
		END

	lastarg = argv[argc-1];
    	
	IF (argc < 3) optionerr = TRUE;
	c = lastarg[strlen(lastarg)-1];
	IF (
	    (argc > 3)
	    &&
	    (c != '/' && c != ':' && strcmp(lastarg,"."))
	   ) optionerr = TRUE;
	
	IF (optionerr) 
	    BEGIN
		printf("Usages: COPY [u/][d:]filename [u/][d:]newname [-V]\n");
		printf("        COPY [u/][d:]filename u/ [-V]\n");
		printf("        COPY [u/][d:]filename d: [-V]\n");
		printf("        COPY [u/][d:]filename .  [-V]\n");
		printf("        COPY [u/][d:]filename u/d:  [-V]\n");
		printf("        COPY <afn>  u/ [-V]\n");
		printf("        COPY <afn>  d: [-V]\n");
		printf("        COPY <afn>  .  [-V]\n");
		printf("        COPY <afn>  u/d: [-V]\n");
		printf("-V      Verify file copy.\n");
		printf("-O      Overwrite existing file.\n");
		exit(0);
	    END
    
/*********************************************************************
***                  END OF ARGUMENT PROCESSING                    ***
*********************************************************************/
    argc--;
    wildexp(&argc,&argv);

    cur_disk = 'A' + bdos(25);

    IF (verify)
    vbuf = sbrk(SECSIZ);

    corebuf = sbrk(SECSIZ);
    FOR (bufsize = SECSIZ; sbrk(SECSIZ) != ERROR; bufsize += SECSIZ)
    ;
    bufsects = bufsize / SECSIZ;


    FOR (loopcnt = 1; loopcnt < argc; loopcnt++)
      LOOP
	IF ((fd1 = open(argv[loopcnt],0)) == ERROR) 
	  THEN
	    printf("\nCOPY: Can't open %s\n",argv[loopcnt]);
	    CONTINUE;  	/* go on to next one anyway. */
	  ENDIF

	strcpy(destname,lastarg);  	/* create output filename */

	IF ( (c = destname[strlen(destname) - 1])=='/' || c == ':' ||
	!strcmp(destname,"."))
	  THEN
	    IF (!strcmp(destname,"."))
	    *destname = '\0';
	    FOR (i = strlen(argv[loopcnt]) - 1; i >= 0; i--)
	    IF (argv[loopcnt][i] == '/' || argv[loopcnt][i] == ':')
	    BREAK;
	    strcat(destname,&argv[loopcnt][i+1]);
	  ENDIF

	IF (!overwrite)
	THEN
	IF ((fd2 = open(destname,0)) != ERROR)  /* if file exists */
	    THEN
		close(fd2);
		printf("\nCOPY: File  %s exists. Overwrite (Y or N)? :"
		    ,destname);
		IF (toupper(getchar()) != 'Y')
		    THEN
			printf("\nCOPY: File %s not copied.\n"
			    ,destname);
			CONTINUE;
		    ENDIF
		printf("\n");
	    ENDIF
	ENDIF

	IF ((fd2 = creat(destname)) == ERROR) 
	  THEN
	    printf("\nCOPY: Can't create %s\n",destname);
	    printf("COPY: Assuming out of directory space and aborting.\n");
	    exit();
	  ENDIF

	putchar('\n');
	printf("COPY: Copying %s to %s...",argv[loopcnt],destname);

	WHILE (1)
	  LOOP
	    IF (kbhit()) getchar();
	    IF (!(i = read(fd1,corebuf,bufsects))) BREAK;
	    IF (i == ERROR)
	      THEN
		printf("\nCOPY: Read error: tell(fd1) = %d, \"%s\"\n",
		tell(fd1), errmsg(errno()));
		BREAK;
	      ENDIF

	    IF (kbhit()) getchar();
	    IF (write(fd2,corebuf,i) != i) 
	      THEN
		printf("\nCOPY: Write error: %s\n",errmsg(errno()));
		exit();
	      ENDIF
	    IF (verify)
	      THEN
		puts("[v] ");
		seek(fd2, -i, 1);
		FOR (j = 0, k = corebuf; j < i; j++, k += SECSIZ)
		  LOOP
		    IF (read(fd2, vbuf, 1) != 1)
		    printf("\nCOPY: Verify read error on %s\n",
		    destname);
		    IF (memcmp(vbuf, k, SECSIZ))
		    CONTINUE;
		    ELSE
		    printf("\nCOPY: Verify error on %s\n",
		    destname);
		  ENDLOOP
	      ENDIF
	  ENDLOOP

	IF (close(fd2) == ERROR) 
	  THEN
	    printf("\nCOPY: Can't close the output file.\n");
	    ;
	    exit();
	  ENDIF
	fabort(fd1);
      ENDLOOP
  END

/*
	Return true if the string arg is a filename prefixed by "nn/",
	where "nn" is a user number:
*/

FUNCTION INT hasuno(str)
CHAR *str;
  BEGIN
    CHAR c;
    INT sum;

    sum = 0;

    IF (!isdigit(*str)) RETURN FALSE;

    WHILE (isdigit(c = *str++))
    sum = sum * 10 + c - '0';
    RETURN (c == '/') ? (sum >= 0 && sum < 32) : FALSE;
  END


/*********************************************************************
*                           ERROR
*********************************************************************/
FUNCTION error(format)
CHAR *format;
  BEGIN
    FUNCTION void putchar();
    _spr (&format, &putchar);
    putchar('\n');
    exit();
  END
