/*
%CC1 $1.C -X
*/
/*********************************************************************
*                               DIO                                  *
*                Written by Gene Mallory  12/24/83                   *
**********************************************************************
*                  COPYRIGHT 1983 EUGENE H. MALLORY                  *
*********************************************************************/
/*
	Redirection can take the following forms:
	
	    > fid	Output redirection to file
	    < fid	Input rediredtion from file
	    + fid	Output redirection to file and screen
	    ~ fid	Output redirection to file and printer
	    >> fid	Append to file
	    +> fid	Append to file and output to screen
	    ~> fid	Append to file and output to printer
	    > LST:	Output to LST device
	    >> LST:	As above
	    + LST:	Output to LST device and screen
	    > LP1:	Output to LP1 (IOBYTE = 0x80)
	    >> LP1:	As above
	    + LP1:	As above and to screen
	    > LP2:	Output to LP2 (IOBYTE = 0xC0)
	    >> LP2:	As above
	    + LP2:	As above and to screen
	    > OUT:	Output to PUNCH DEVICE
	    >> OUT:	As above
	    + OUT:	As above and to screen
	    ~ OUT:	As above and to printer
	    < IN:	Input from READER device
	    
	    | prog	Pipe output to next program
	    +| prog	As above and to screen
	    ~| prog	As above and to printer
	
	Redirection and pipe character may be separated from
	the file or program name by white space or not.
	
	File specifications may be complete with user number (##) and 
	disk name (d) in the form:
	
			##/d:name
	
	Pipe specification will be first tried as is, if the file 
	does not exist, disk A, user 0 will be tried using the same 
	name.
	
	Control C will abort program and submit file execution.
	When inputing from keyboard use Control C except at
	column 1. (CP/M catches it there.)  If an old version of
	the output file existed, it will be recovered by Control C.
	
	It is permissable to input and output from the same file,
	BUT the input specification must be first and the output
	specification must be second!  otherwise the order and 
	location of the redirection specifications is not
	significant.
	
	    
	
	
	Thus, a C program using redirection has the following form:

	    #include "bdscio.h"		/* standard header file	*/
	    #include "dio.h"		/* directed I/O header	*/
	    
		...				/* other externals, if any */

	    main(argc,argv)
		char **argv;
		BEGIN
		    ...			/* declarations		     */
		    dioinit(&argc,argv)	/* initialize redirection    */
		    wildexp(&argc,&argv)/* wildexp goes here if used */
		    ...			/* body of program	     */
		    dioflush();		/* Must exit here to close   */
		END
*/
/*
	NOTES:

	1. Redirection and pipes work only for TEXT. This mechanism 
	should not be used for binary data.

	2. Use CLINK FOO DIO or L2 FOO DIO to force 
	linking with all the dio routines before the DEFFs are 
	searched.
	
	3. Library routines GETCHAR, PUTCHAR, PRINTF, PUTS are
	caught and redirected.
	
	4. New routines TYPEF, TYPECHAR FORCE OUTPUT TO 
	SCREEN.
	
	5. New routine ERROR, may be used like TYPEF to 
	output a message, close out the DIO operations
	and exit.
	

*/

#INCLUDE "bdscio.h"
#INCLUDE "dio.h"

/*********************************************************************
*                           DIOINIT                                  *
*********************************************************************/
FUNCTION dioinit(argcptr,argvect)
INT *argcptr;
CHAR **argvect;
  BEGIN
    INT i, c0, c1;
    _outptr = STDERR;
    _conline = MAXLINE;
    _conbufp = _conbuf;
    _conbuf[0] = 0;
    _iobyte = bdos(7,0);  	/* save iobyte */
    _screenflag = _pipeflag = _dioinflag = FALSE;
    _printerflag = _appendflag = _rdrflag = FALSE;

    FOR (i=1;i<*argcptr;)
      LOOP
	c0 = argvect[i][0];
	c1 = argvect[i][1];
	IF (c0 == '+') _screenflag = TRUE;
	IF (c0 == '~') _printerflag = TRUE;

	IF ((c0=='+' || c0=='~') && c1 == '|')
	  THEN
	    strshift(argvect[i],1);
	    _dio_pipe(i,argcptr,argvect);
	  ENDIF
	ELSE IF ((c0=='+' || c0=='~') && c1 == '>')
	  THEN
	    _dioappend(i,argcptr,argvect);
	  ENDIF
	ELSE IF (c0 == '>' && c1 == '>')
	  THEN
	    _dioappend(i,argcptr,argvect);
	  ENDIF
	ELSE IF (c0 == '>' || (c0=='+' || c0=='~')) 
	  THEN
	    _dioout(i,argcptr,argvect);
	  ENDIF
	ELSE IF (c0 == '<') 
	  THEN
	    _dioin(i,argcptr,argvect);
	  ENDIF
	ELSE IF (c0 == '|') 
	  THEN
	    _dio_pipe(i,argcptr,argvect);
	  ENDIF
	ELSE 
	  THEN
	    i++;
	  ENDIF
      ENDLOOP
 END

/*********************************************************************
*                           _DIOOUT                                  *
*********************************************************************/
FUNCTION _dioout(n,argcptr,argvect)
INT *argcptr,n;
CHAR **argvect;
  BEGIN
    CHAR backupfile[20];
    _outptr = _dobuf;

    IF (argvect[n][0] == '+') _screenflag = TRUE;

    IF (strlen(argvect[n]) == 1) shiftarg(n,argcptr,argvect);
    ELSE strshift(argvect[n],1);

    IF (!strcmp(argvect[n],"LST:")) _outptr = STDLST;
    ELSE IF (!strcmp(argvect[n],"OUT:")) _outptr = STDPUN;
    ELSE IF (!strcmp(argvect[n],"LP1:")) 
      THEN
	_outptr = STDLST;
	bdos(8,(_iobyte&0x3f)|0x80);
      ENDIF
    ELSE IF (!strcmp(argvect[n],"LP2:"))
      THEN
	_outptr = STDLST;
	bdos(8,(_iobyte&0x3f)|0xC0);
      ENDIF
    ELSE 
      THEN
	strcpy(_oldfile,argvect[n]);
	getprefix(backupfile,_oldfile);
	strcat(backupfile,"BACKUP.$$$");
	unlink(backupfile);
	IF (DIOIN && (strcmp(argvect[n],_infile) == 0)) 
	BEGIN
	fclose(_dibuf);
	IF (rename(argvect[n],backupfile) == -1)
	    {
	    fprintf(STDERR,"DIO: Failed to rename %s to %s.\n",
	    argvect[n],backupfile);
	    dioexit();
	    }
	IF (fopen(backupfile,_dibuf) == ERROR)
	    THEN
		fprintf(STDERR,"DIO: Unable to open file %s for input.\n"
		,backupfile);
		fprintf(STDERR,"DIO: Error caused by %s.\n",errmsg(errno()));
		dioexit();
	    ENDIF
	END
	ELSE {
	rename(argvect[n],backupfile);
	    }
	IF (fcreat(argvect[n],_dobuf) == ERROR)
	  THEN
	    fprintf(STDERR,"DIO: Unable to open file %s for output.\n"
	    ,argvect[n]);
	    fprintf(STDERR,"DIO: Error caused by %s.\n",errmsg(errno()));
	    dioexit();
	  ENDIF
      ENDIF
    shiftarg(n,argcptr,argvect);
  END

/*********************************************************************
*                           _DIO_PIPE
*********************************************************************/
FUNCTION _dio_pipe(n,argcptr,argvect)
INT *argcptr,n;
CHAR **argvect;
  BEGIN
    IF (strlen(argvect[n]) == 1) shiftarg(n,argcptr,argvect);
    ELSE strshift(argvect[n],1);

    _pipeflag = 1;
    _outptr = _dobuf;

    _pargv = &argvect[n];
    _pargc = *argcptr-n;
    *argcptr = n;

    IF (fcreat("TEMPOUT.$$$",_dobuf) == ERROR)
      THEN
	fprintf(STDERR,"DIO: Unable to open file TEMPOUT.$$$ for pipe.\n");
	fprintf(STDERR,"DIO: Error caused by %s.\n",errmsg(errno()));
	dioexit();
      ENDIF
  END

/*********************************************************************
*                           _DIOAPPEND
*********************************************************************/
FUNCTION _dioappend(n,argcptr,argvect)
INT *argcptr,n;
CHAR **argvect;
  BEGIN
    _appendflag = TRUE;
    _outptr = _dobuf;

    IF (argvect[n][0] == '+') _screenflag = TRUE;

    IF (strlen(argvect[n]) == 2) shiftarg(n,argcptr,argvect);
    ELSE strshift(argvect[n],2);

    IF (!strcmp(argvect[n],"LST:")) _outptr = STDLST;
    ELSE IF (!strcmp(argvect[n],"OUT:")) _outptr = STDPUN;
    ELSE IF (!strcmp(argvect[n],"LP1:")) 
      THEN
	_outptr = STDLST;
	bdos(8,(_iobyte&0x3f)|0x80);
      ENDIF
    ELSE IF (!strcmp(argvect[n],"LP2:") )
      THEN
	_outptr = STDLST;
	bdos(8,(_iobyte&0x3f)|0xC0);
      ENDIF
    ELSE IF (fappend(argvect[n],_dobuf) == ERROR)
      THEN
	fprintf(STDERR,"DIO: Unable to open file %s for append.\n"
	,argvect[n]);
	fprintf(STDERR,"DIO: Error caused by %s.\n",errmsg(errno()));
	dioexit();
      ENDIF
    shiftarg(n,argcptr,argvect);
  END

/*********************************************************************
*                           _DIOIN
*********************************************************************/
FUNCTION _dioin(n,argcptr,argvect)
INT *argcptr,n;
CHAR **argvect;
  BEGIN
    IF (strlen(argvect[n]) == 1) shiftarg(n,argcptr,argvect);
    ELSE strshift(argvect[n],1);

    _dioinflag = 1;

    IF (!strcmp(argvect[n],"IN:")) _rdrflag = TRUE;
    ELSE IF (fopen(argvect[n],_dibuf) == ERROR)
      THEN
	fprintf(STDERR,"DIO: Unable to open file %s for input.\n"
	,argvect[n]);
	fprintf(STDERR,"DIO: Error caused by %s.\n",errmsg(errno()));
	dioexit();
      ENDIF
    strcpy(_infile,argvect[n]);
    shiftarg(n,argcptr,argvect);
  END

/*********************************************************************
*                           SHIFTARG
*********************************************************************/
FUNCTION shiftarg(n,argcptr,argvect)
INT n,*argcptr;
CHAR **argvect;
  BEGIN
    INT i;
    *argcptr = *argcptr-1;
    FOR (i=n;i<*argcptr;i++)
      LOOP
	argvect[i] = argvect[i+1];
      ENDLOOP
  END

/*********************************************************************
*                           STRSHIFT
*********************************************************************/
FUNCTION strshift(string,n)
INT n;
CHAR *string;
  BEGIN
    INT i;
    FOR (i=n;string[i];i++) string[i-n] = string[i];
    string[i-n] = 0;
  END

/*********************************************************************
*                           DIOFLUSH
*********************************************************************/
FUNCTION dioflush()
  BEGIN
    INT fid;
    CHAR fname[20];
    IF (_dioinflag && !_rdrflag) 
      THEN
	fabort(_dibuf);
      ENDIF
    IF (_outptr == _dobuf)
      THEN
	putc(CPMEOF,_dobuf);
	fclose(_dobuf);
      ENDIF
    bdos(8,_iobyte);  	/* restore iobyte */
    IF (_pipeflag)
      THEN
	unlink("TEMPIN.$$$");
	rename("TEMPOUT.$$$","TEMPIN.$$$");
	strcpy(fname,_pargv[0]);
	strcat(fname,".com");
	IF ((fid = open(fname,READ)) == ERROR)
	  THEN
	    close(fid);
	    strcpy(fname,"0/A:");
	    strcat(fname,_pargv[0]);
	    strcat(fname,".com");
	  ENDIF
	close(fid);
	_pargv[0] = "<TEMPIN.$$$";
	_pargv[_pargc] = 0;
	IF (execv(fname,_pargv) == -1)
	  THEN
	    fprintf(STDERR,"DIO: Unable to _pipe to %s.\n",fname);
	    dioexit();
	  ENDIF
      ENDIF
    ELSE
      THEN
	unlink("TEMPOUT.$$$");
	unlink("TEMPIN.$$$");
	unlink("BACKUP.$$$");
      ENDIF
    exit();
  END

/*********************************************************************
*                           PUTCHAR
*********************************************************************/
FUNCTION INT putchar(c)
CHAR c;
  BEGIN
    IF (_screenflag)
      THEN
	IF (c == '\n') putc('\r',STDERR);
	putc(c,STDERR);
      ENDIF
    IF (_printerflag)
      THEN
	IF (c == '\n') putc('\r',STDLST);
	putc(c,STDLST);
      ENDIF
    IF (c == '\n') 
      THEN
	IF (bdos(11) && bdos(1) == 3) 
	  THEN
	    typef("\nDIO: Control C intercepted.\n");
	    dioexit();
	  ENDIF
	putc('\r',_outptr);
      ENDIF
    RETURN putc(c,_outptr);
  END

/*********************************************************************
*                           GETCHAR
*********************************************************************/
FUNCTION INT getchar()
  BEGIN
    INT c,c2;
    IF (_dioinflag && !_rdrflag) 
      THEN
	c = getc(_dibuf);
	IF (c == '\r')
	  THEN
	    IF (bdos(11) && bdos(1) == 3) 
	      THEN
		typef("\nDIO: Control C intercepted.\n");
		dioexit();
	      ENDIF
	    IF ((c2 = getc(_dibuf)) == '\n') c = c2;
	    ELSE ungetc(c2,_dibuf);
	  ENDIF
      ENDIF
    ELSE IF (_rdrflag) 
      THEN
	c = getc(STDRDR);
	IF (c == '\n') c = getc(STDRDR);
	IF (c == '\r') 
	  THEN
	    IF (bdos(11) && bdos(1) == 3) 
	      THEN
		typef("\nDIO: Control C intercepted.\n");
		dioexit();
	      ENDIF
	    c = '\n';
	  ENDIF
      ENDIF
    ELSE
      THEN
	/* from the BDOS (using "gets"), & insert newline:  */
	IF (!*_conbufp) 
	  THEN
	    typechar('*');
	    bdos(10,&_conline);
	    bdos(2,'\n');  	/* Echo lf after line. */
	    _conbuf[_concount] = '\n';
	    _conbuf[_concount+1]=0;
	    _conbufp = _conbuf;
	  ENDIF
	c = *_conbufp++;
	IF (c == 3) 
	  THEN
	    typef("\nDIO: Control C intercepted.\n");
	    dioexit();
	  ENDIF
      ENDIF
    IF (c == CPMEOF) RETURN EOF;
    RETURN c;
  END

/****************************************************************/
/*	DIO DIRECTED GET STRING FUNCTION                        */
/*      Returns 1 at end of file, 0 normally.                   */
/****************************************************************/
FUNCTION INT getstring(cptr)
CHAR *cptr;
  BEGIN
    INT c;
    WHILE (TRUE)
      LOOP
	c = getchar();
	IF (c == EOF || c == CPMEOF) 
	  THEN
	    *cptr++ = 0;
	    RETURN 1;
	  ENDIF
	IF (c == '\n')
	  THEN
	    *cptr++ = c;
	    *cptr++ = 0;
	    RETURN 0;
	  ENDIF
	*cptr++ = c;
      ENDLOOP
  END

/*********************************************************************
*                           TYPECHAR
*********************************************************************/
FUNCTION INT typechar(c)
CHAR c;
  BEGIN
    RETURN putc(c,STDERR);
  END

/*********************************************************************
*                           TYPEF
*********************************************************************/
FUNCTION typef(format)
CHAR *format;
  BEGIN
    RETURN _spr (&format, &typechar);
  END

/*********************************************************************
*                           ERROR
*********************************************************************/
FUNCTION error(format)
CHAR *format;
  BEGIN
    _spr (&format, &typechar);
    typechar('\n');
    dioexit();
  END

/*********************************************************************
*                           DIOEXIT
*********************************************************************/
FUNCTION dioexit()
  BEGIN
    CHAR backupfile[20];
    IF ((_outptr == _dobuf) && !_pipeflag && !_appendflag)
      THEN
	putc(CPMEOF,_dobuf);
	fclose(_dobuf);
	getprefix(backupfile,_oldfile);
	strcat(backupfile,"BACKUP.$$$");
	unlink(_oldfile);
	IF (rename(backupfile,_oldfile) != ERROR)
	  THEN
	    typef("DIO: Oldfile, %s, recovered from %s.\n"
	    ,_oldfile,backupfile);
	  ENDIF
      ENDIF
    unlink("TEMPOUT.$$$");
    unlink("TEMPIN.$$$");
    unlink("0/A:$$$.SUB");
    typef("DIO: Aborting job.\n");
    exit();
  END

/*********************************************************************
*                           GETPREFIX
*********************************************************************/
FUNCTION getprefix(s1,s2)
CHAR *s1,*s2;
  BEGIN
    CHAR *s1save,c;
    s1save = s1;
    WHILE (c = *s2++)
      LOOP
	*s1++ = c;
	IF (c == '/') 
	  THEN
	    s1save = s1;
	  ENDIF
	IF (c == ':') 
	  THEN
	    s1save = s1;
	    BREAK;
	  ENDIF
      ENDLOOP
    *s1save = 0;
    RETURN;
  END
